#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""


"""

import os
import pickle
import numpy as np
from numpy.fft import rfft, rfftfreq, irfft
from scipy.interpolate import LinearNDInterpolator
from scipy.interpolate import NearestNDInterpolator
from scipy.interpolate import CloughTocher2DInterpolator
from scipy.integrate import quad
from scipy.stats import pearsonr
from obspy import Trace, Stream
from obspy.geodetics import gps2dist_azimuth
from obspy import UTCDateTime as UTC
from seislib.exceptions import TimeSpanException


def gc_distance(lat1, lon1, lat2, lon2):
    """ 
    Calculates the great circle distance (in m) between coordinate points 
    (in degrees). This function calls directly the obspy `gps2dist_azimuth`,
    it only extends its functionality through the `numpy.vectorize` decorator.
    
    Parameters
    ----------
    lat1, lon1, lat2, lon2 : float or array-like of shape (n,)
        Coordinates of the points on the Earth's surface, in degrees.        
    
    Returns
    -------
    Great-circle distance (in m) 
        If the input is an array (or list) of coordinates, an array of 
        distances is returned
    """
    func = np.vectorize(gps2dist_azimuth)
    return func(lat1, lon1, lat2, lon2)[0]


def azimuth_backazimuth(lat1, lon1, lat2, lon2):
    """ 
    Calculates the azimuth and backazimuth (in degrees) between coordinate 
    points (in degrees). This function calls directly the obspy 
    `gps2dist_azimuth`, it only extends its functionality through the 
    numpy.vectorize decorator
    
    Parameters
    ----------
    lat1, lon1, lat2, lon2 : float or array-like of shape (n,)
        Coordinates of the points on the Earth's surface, in degrees.               
    
    Returns
    -------
    tuple of shape (2,) or ndarray of shape (n, 2)
        The columns consist of azimuth azimuth and backazimuth    
    """
    func = np.vectorize(gps2dist_azimuth)
    return func(lat1, lon1, lat2, lon2)[1:]


def adapt_timespan(st1, st2):
    """
    Slices all traces from the two input streams to the overlapping timerange.
    Then returns a copy of the sliced streams.
        
    Parameters
    ----------
    st1, st2 : obspy.Stream, obspy.Trace

    Returns
    -------
    st1, st2 : obspy.Stream, obspy.Trace
        Obspy stream or trace depending on the input. The original input is 
        not permanently modified (a copy is returned)

    Raises
    ------
    TimeSpanException
        If no overlap is available

    Notes
    -----
    The maximum precision achieved by this function is governed by the
    samling rate. If sub-sample precision is required, consider using 
    :meth:`adapt_timespan_interpolate`
    """
    is_trace = False
    if isinstance(st1, Trace) or isinstance(st2, Trace):
        is_trace = True
    st1 = Stream(st1) if isinstance(st1, Trace) else st1
    st2 = Stream(st2) if isinstance(st2, Trace) else st2
    
    # this has to be done twice, because otherwise there sometimes occurs a 1s timeshift
    for adapt in range(2):
        starttime = max([tr.stats.starttime for tr in st1]
                        + [tr.stats.starttime for tr in st2])   
        endtime = min([tr.stats.endtime for tr in st1]
                      + [tr.stats.endtime for tr in st2])
        if starttime >= endtime:
            raise TimeSpanException(st1, st2)
            
        st1 = st1.slice(starttime, endtime)
        st2 = st2.slice(starttime, endtime)
        for tr in st1:
            tr.stats.starttime = starttime
        for tr in st2:
            tr.stats.starttime = starttime
            
    return (st1, st2) if not is_trace else (st1[0], st2[0])


def adapt_timespan_interpolate(st1, st2, min_overlap=0):
    """
    Slices all traces from the two input streams to the overlapping timerange.
    Then returns a copy of the sliced streams. If the starttime of the sliced 
    traces do not fit exactly (because of sub-sample time shifts), the traces 
    are interpolated to remove the time shift.       
    
    Parameters
    ----------
    st1, st2 : obspy.Stream, obspy.Trace

    Returns
    -------
    st1_out, st2_out : obspy.Stream, obspy.Trace
        Obspy stream or trace depending on the input. The original input
        is not permanently modified (a copy is returned)

    Raises
    ------
    TimeSpanException
        If no overlap is available

    Notes
    -----
    Interpolation can require a relatively long time depending on the
    size of the sliced stream. If speed is preferred to (sub-sample) 
    precision, consider using :meth:`adapt_timespan`
    """
    def slice_streams(st1, st2, starttime, endtime):
        st1_out = st1.slice(starttime, endtime)
        st2_out = st2.slice(starttime, endtime)
        return st1_out, st2_out
    
    is_trace = False
    if isinstance(st1, Trace) or isinstance(st2, Trace):
        is_trace = True
    st1 = Stream(st1) if isinstance(st1, Trace) else st1
    st2 = Stream(st2) if isinstance(st2, Trace) else st2
    starttime = max([tr.stats.starttime for tr in st1]
                    + [tr.stats.starttime for tr in st2])
    endtime = min([tr.stats.endtime for tr in st1]
                  + [tr.stats.endtime for tr in st2])
    
    if starttime > endtime:
        raise TimeSpanException(st1, st2)
    
    sr1 = st1[0].stats.sampling_rate
    sr2 = st2[0].stats.sampling_rate
    st1_out, st2_out = slice_streams(st1, st2, starttime, endtime)
    
    starttimes = set([tr.stats.starttime.timestamp for tr in st1_out]
                     + [tr.stats.starttime.timestamp for tr in st2_out])
    if len(starttimes) > 1:
        starttime = UTC(max(starttimes))
        st1_out.interpolate(sr1, starttime=starttime)
        st2_out.interpolate(sr2, starttime=starttime)
        endtime = min([tr.stats.endtime for tr in st1_out]
                      + [tr.stats.endtime for tr in st2_out])
        st1_out, st2_out = slice_streams(st1_out, st2_out, starttime, endtime)

    return (st1_out, st2_out) if not is_trace else (st1_out[0], st2_out[0])
    

def adapt_sampling_rate(st1, st2):
    """ 
    If the input streams (or traces) have different sampling rates, the one
    characterized by the largest sampling rate is downsampled to the sampling
    rate of the other stream (or trace).
    
    The downsampling is carried out via the :meth:`resample`, which modifies 
    the input streams in place.
    
    Parameters
    ----------
    st1, st2 : obspy.Stream, obspy.Trace
    
    Returns
    -------
    st1, st2 : obspy.Stream, obspy.Trace
        Obspy stream or trace depending on the input. The input is permanently 
        modified
    """
    is_trace = False
    if isinstance(st1, Trace) or isinstance(st2, Trace):
        is_trace = True
    st1 = Stream(st1) if isinstance(st1, Trace) else st1
    st2 = Stream(st2) if isinstance(st2, Trace) else st2
    fs1, fs2 = st1[0].stats.sampling_rate, st2[0].stats.sampling_rate  
    if fs1 < fs2:
        st2 = resample(st2, fs1)
    elif fs2 < fs1:
        st1 = resample(st1, fs2)
    return (st1, st2) if not is_trace else (st1[0], st2[0])


def resample(x, fs):
    """ 
    If the input streams (or traces) have different sampling rates, the one
    characterized by the largest sampling rate is downsampled to the sampling
    rate of the other stream (or trace).
    
    Parameters
    ----------
    st1, st2 : obspy.Stream, obspy.Trace
    

    Returns
    -------
    st1, st2 : obspy.Stream, obspy.Trace
        Obspy stream or trace depending on the input. The input is permanently 
        modified
    """
    nyquist_f = fs/2 - (fs/2)*0.01
    try:
        x.filter('lowpass', freq=nyquist_f, corners=4, zerophase=True)
    except ValueError: 
        pass # when fs > sampling_rate(x), filtering is not needed
    x.interpolate(sampling_rate=fs, method="weighted_average_slopes")
    return x



def bandpass_gaussian(data, dt, period, alpha):
    """ Gaussian filter of real-valued data carried out in the frequency domain
    
    The bandpass filter is carried out with a Gaussian filter centered at 
    `period`, whose width is controlled by `alpha`::

        exp(-alpha * ((f-f0)/f0)**2)

    where f is frequency and f0 = 1 / period. 
    
    Parameters
    ----------
    data : ndarray of shape (n,)
        Real-valued data to be filtered

    dt : float
        Time sampling interval of the data

    period : float
        Central period, around which the (tight) bandapass filter is carried out

    alpha : float
        Parameter that controls the width of the Gaussian filter
        
    Returns
    -------
    numpy.ndarray of shape (n,) 
        Filtered data
    """
    ft = rfft(data)
    freq = rfftfreq(len(data), d=dt)
    f0 = 1.0 / period
    ft *= np.exp( -alpha * ((freq-f0) / f0)**2 )
    return irfft(ft, n=len(data))


def zeropad(tr, starttime, endtime):
    """ 
    Zeropads an `obspy.Trace` so as to cover the time window 
    specified by `starttime`'and `endtime`
    
    Parameters
    ----------
    tr : obspy.Trace

    starttime, endtime : obspy.UTCDateTime
    
    Returns
    -------
    trace : obspy.Trace
        Zeropadded copy of the input trace.
    """
    trace = Trace()
    for key, value in tr.stats.items():
        if key not in ['endtime', 'npts']:
            trace.stats[key] = value
    fs = tr.stats.sampling_rate
    samples_before = int((tr.stats.starttime - starttime) * fs)
    samples_after = int((endtime - tr.stats.endtime) * fs)
    data = tr.data
    if samples_before > 0:
        trace.stats.starttime = tr.stats.starttime - ((samples_before+1) / fs)
        data = np.concatenate((np.zeros(samples_before+1), data))
    if samples_after > 0:
        data = np.concatenate((data, np.zeros(samples_after+1)))
    
    trace.data = data
    return trace
   

def rotate(r, t, dphi):
    """
    Rotation of radial and transverse component of the seismogram by a specified
    angle, following the obspy signs convention.
    
    Parameters
    ----------
    r, t : numpy.ndarray
        Radial (r) and transverse (t) components
    dphi : float
        Angle in degrees
    
    Returns
    -------
    rnew, tnew : numpy.ndarray
        Rotated components
    """
    rnew = -t*np.sin(np.radians((dphi+180)%360)) - r*np.cos(np.radians((dphi+180)%360))
    tnew = -t*np.cos(np.radians((dphi+180)%360)) + r*np.sin(np.radians((dphi+180)%360))
    return rnew, tnew
                

def rotate_stream(st, **kwargs):
    """        
    The method calls the `obspy.Stream.rotate` method, which sometimes runs into
    errors if differences are present among the starttimes and/or endtimes of 
    the traces constituting the stream. These are prevented by slicing the 
    stream to a common time window and (if necessary) interpolating it so as to 
    avoid sub-sample differences.
    
    Parameters
    ----------
    st : obspy.Stream

    **kwargs
        Optional arguments passed to `obspy.Stream.rotate`
    
    Returns
    -------
    st : obspy.Stream
        Rotated copy of the input Stream
    """
    def starttime_and_endtime(st):
        starttime = max([tr.stats.starttime for tr in st])
        endtime = min([tr.stats.endtime for tr in st])
        return starttime, endtime
    
    try:
        starttime, endtime = starttime_and_endtime(st)
        st = st.slice(starttime, endtime)
        st = st.rotate(**kwargs)
    except ValueError:
        starttime, endtime = starttime_and_endtime(st)
        st.interpolate(sampling_rate=st[0].stats.sampling_rate, 
                       starttime=starttime)
        starttime, endtime = starttime_and_endtime(st)
        st = st.slice(starttime, endtime)
        st = st.rotate(**kwargs)
    return st


def running_mean(x, N):
    """ Moving average
    
    Parameters
    ----------
    x : ndarray of shape (m,)
        Data vector

    N : int
        Controls the extent of the smoothing (larger values correspond to larger
        smoothing)
    
    Returns
    -------
    runmean : ndarray of shape (m,)
        Smoothed input
    
    Notes
    -----
    This is a simple implementation of a moving average. More sofisticated 
    functions can be found, e.g., in `scipy.signal.savgol_filter` or 
    `scipy.ndimage.filters.uniform_filter1d`
    """
    if N%2 == 0:
        N+=1
    idx0 = int((N-1)/2)
    runmean = np.zeros(len(x))
    cumsum = np.cumsum(np.insert(x, 0, 0)) 
    runmean[idx0:-idx0] = (cumsum[N:] - cumsum[:-N]) / N
    for i in range(idx0):
        runmean[i] = np.mean(x[:2*i+1])
        runmean[-i-1] = np.mean(x[-2*i-1:])
    return runmean


def scatter_to_mesh(lats, lons, c, mesh, method='linear'):
    """ Translates scattered data into a seislib mesh
    
    Parameters
    ----------
    lats, lons : ndarray (n,)
        Coordinates of the scattered data

    c : ndarray (n,)
        Values of the scattered data

    mesh : ndarray (m, 4)
        mesh in SeisLib's format, where the four columns correspond 
        to the boundaries of each pixel, i.e., lat1, lat2, lon1, lon2

    method : {'linear', 'nearest', 'cubic'}
        Interpolation method. Supported: 'linear' (default), 'nearest', 
        and 'cubic'. The three methods call `LinearNDInterpolator`, 
        `NearestNDInterpolator`, and `CloughTocher2DInterpolator` of the 
        `scipy.interpolate` module, respectively.
    
    Returns
    -------
    1-D ndarray containing the `c` values interpolated on `mesh`

    Raises
    ------
    NotImplementedError
        If `method` is not supported
    """
    if method == 'linear':
        interpolator = LinearNDInterpolator(np.column_stack((lons, lats)), c)
    elif method == 'nearest':
        interpolator = NearestNDInterpolator(np.column_stack((lons, lats)), c)
    elif method == 'cubic':
        interpolator = CloughTocher2DInterpolator(np.column_stack((lons, lats)), c)
    else:
        msg = '`%s` interpolation not supported. Supported '%method
        msg += 'methods are `linear`, `nearest`, and `cubic`'
        raise NotImplementedError(msg)
    mesh_central_coords = np.column_stack(((mesh[:,2]+mesh[:,3]) / 2, 
                                           (mesh[:,0]+mesh[:,1]) / 2))
    return interpolator(mesh_central_coords)
    
    
def pearson_corrcoef(v1, v2):
    """ Pearson coerrelation coefficient between two vectors
    
    Parameters
    ----------
    v1, v2 : lists or ndarrays (n,)
        
    
    Returns
    -------
    corrcoeff : float
        Pearson correlation coefficient

    pvalue : float
    

    Raises
    ------
    ValueError
        If v1 and v2 have different shapes

    Notes
    -----
    This function calls `scipy.stats.pearsonr`, bu handles the presence
    of `nan` values.
    """
    if v1.shape != v2.shape:
        raise ValueError('Shapes %s and %s are inconsistent'%(v1.shape, v2.shape))
    notnan = np.intersect1d(np.flatnonzero(~np.isnan(v1)), 
                            np.flatnonzero(~np.isnan(v2)))
    return pearsonr(v1[notnan], v2[notnan])


@np.vectorize
def gaussian(x, mu, sigma):
    """ Gaussian function
    
    Parameters
    ----------
    x : float or ndarray
        Indipendent variable
        
    mu : float
        Mean of the Gaussian
        
    sigma : float
        Standard deviation of the Gaussian
        
    
    Returns
    -------
    float or ndarray
        Gaussian evaluated at x
    """
    return 1 / (sigma * np.sqrt(2*np.pi)) * np.exp(-1/2 * ((x-mu)/sigma)**2)
   

@np.vectorize
def skewed_normal(x, mu, sigma, skewness):   
    """ Skewed Normal distribution
    
    Parameters
    ----------
    x : float or ndarray
        Indipendent variable
        
    mu : float
        Mean of the resulting Skewed Normal
        
    sigma : float
        Standard deviation of the Skewed Normal
        
    skewness : float
        Parameter regulating the skewness of the function. The function is
        right-skewed if `skewness`>0, and left-skewed if `skewness`<0

    Returns
    -------
    float or ndarray
        Skewed Normal distribution evaluated at x
    """    
    loc = mu - (np.sqrt(2 / np.pi)) * (sigma*skewness / np.sqrt(1+skewness**2))
    integrand = lambda t: np.exp(-t**2 / 2)
    const = 1 / (sigma*np.pi) * np.exp(-(x-loc)**2 / (2*sigma**2))
    suplim = skewness*((x - loc)/sigma)
    return const * quad(integrand, -np.inf, suplim)[0]


def next_power_of_2(x):  
    """ Closest power of two larger or equal to x
    
    Parameters
    ----------
    x : int
    
    
    Returns
    -------
    int
    """
    return 1 if x==0 else 2**(x - 1).bit_length()


def load_pickle(path):
    """ Loads a .pickle file
    
    Parameters
    ----------
    path : str
        Absolute path to the file
    
    Returns
    -------
    Object contained in the .pickle file
    """
    with open(path, 'rb') as f:
        return pickle.load(f) 


def save_pickle(file, obj):
    """ Saves an object to a .pickle file
    
    Parameters
    ----------
    file : str
        Absolute path to the resulting file
    
    obj : python object 
        Object to be saved (see documentation on the pickle module to know
        more on which Python objects can be stored into .pickle files)
    """
    with open(file, 'wb') as f:
        pickle.dump(obj, f)


def remove_file(file):
    """ Removes a file from disk, handling eventual exceptions
    
    Parameters
    ----------
    file : str
        Absolute path to the file to be removed
    """
    try:
        os.remove(file)
    except FileNotFoundError:
        pass






r"""
==============================================
Seismic Tomography (:mod:`seislib.tomography`)
==============================================

This module allows for inverting inter-station (or epicenter-station) 
velocity measurements for surface-wave velocity maps. SeisLib implements a 
least-square inversion scheme based on ray theory [1]_, i.e., it assumes that 
the surface waves travel, from a given point on the Earth's surface to 
another, without deviating from the great-circle path connecting them. 

By default, to discretize the Earth's surface, 
:class:`seislib.tomography._tomography.SeismicTomography` 
implements an equal-area grid (see :class:`seislib.tomography._grid.EqualAreaGrid`). 
Alternatively, a regular grid can be employed (see 
:class:`seislib.tomography._grid.EqualAreaGrid`). (The latter is particularly 
suited to tomographic applications at local scale, where the use of 
equal-area parameterizations does not have clear advantages.) The grid of choice 
can be refined, in the areas characterized by a relatively high density of
available rays, an arbitrary of times. The refinement is carried out by splitting 
a given block of the parameterization into four sub-blocks.

.. [1] Boschi & Dziewonski 1999, High-and low-resolution images of the Earth's mantle: 
    Implications of different approaches to tomographic modeling, JGR
"""
from ._grid import *
from ._tomography import *


__all__ = [
    'EqualAreaGrid',
    'RegularGrid',
    'SeismicTomography'
    ]




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Inter-Station Dispersion Curves
=============================== 

The below class provides automated routines to calculate
Rayleigh and Love inter-station dispersion curves based on
the continuous seismograms recorded at pairs of receivers.

"""

import os
import itertools as it
import numpy as np
from scipy.interpolate import interp1d
from obspy import read
from obspy.geodetics.base import gps2dist_azimuth
from obspy.io.sac.util import SacIOError
import matplotlib.pyplot as plt
from seislib.utils import rotate_stream, load_pickle, save_pickle, remove_file
from seislib.an import noisecorr, velocity_filter, extract_dispcurve
from seislib.plotting import plot_stations
from seislib.tomography import SeismicTomography


def _read(*args, verbose=True, **kwargs):
    try:
        return read(*args, **kwargs)
    except SacIOError:
        if verbose:
            print('SacIOError', *args)
        

class AmbientNoiseVelocity:
    """ 
    Class to obtain surface-wave phase velocities out of continuous seismograms,
    via cross-correlation of seismic ambient noise. 
    
    Parameters
    ----------
    src : str
        Absolute path to the directory containing the files associated with 
        the continuous seismograms
        
    savedir : str, optional
        Absolute path to the directory where the results are saved. If not
        provided, its value is set to the parent directory of `src` (i.e.,
        $src/..). All the results will be saved in the directory
        $savedir/an_velocity/$component (see the `component` parameter)
        
    component : {'Z', 'R', 'T'}
        Either 'Z', 'R', (corresponding to vertically and radially  polarized 
        Rayleigh waves, respectively), or 'T' (for Love waves)
        
    verbose : bool
        Whether or not information on progress is printed in the console

        
    Attributes
    ----------
    src : str
        Absolute path to the directory containing the files associated with the
        continuous seismograms
        
    savedir : str
        Absolute path to the directory where the results are saved
        
    component : {'Z', 'R', 'T'}
        Either 'Z', 'R', or 'T'
        
    files : list
        List of files to be processed to calculate the dispersion curves
        
    verbose : bool
        Whether or not information on progress is printed in the console
        
    stations : dict
        Station coordinates that can be employed to retrieve dispersion 
        curves (can be accessed after calling the method :meth:`prepare_data`)

    
    Examples
    --------
    The following will calculate the dispersion curves for a set of continuous
    seismograms stored in the directory src='/path/to/data'. The data have been
    downloaded using :class:seislib.an.an_downloader.ANDownloader. We extract
    dispersion curves for vertically-polarized Rayleigh waves (on the vertical 
    - Z - component).
    
    >>> from seislib.an import AmbientNoiseVelocity
    >>> an = AmbientNoiseVelocity(
    ...         src=src,
    ...         component='Z')
    >>> print(an)    
    RAYLEIGH-WAVE VELOCITY FROM SEISMIC AMBIENT NOISE
    ==================================================
    RECEIVERS: X
    STATION PAIRS AVAILABLE: Y
    DISPERSION CURVES RETRIEVED: 0
    ==================================================
    SOURCE DIR: /path/to/data
    SAVE DIR: /path/to/an_velocity/Z

    In the above printed information X and Y are integers and depend on your data.
        
    Before extracting the dispersion curves, we need to run::
        
        an.prepare_data()
    
    which will extract information on each seismic receiver from the header of 
    the sac files. These include (i) station coordinates and (ii) the time window 
    spanned by the associated seismogram. The information is saved into two 
    separate files, i.e., /path/to/an_velocity/Z/stations.pickle and 
    /path/to/an_velocity/Z/timespans.pickle. These include (i) stations 
    coordinates and (ii) the time window spanned by the associated seismograms.
    
    .. note:: 
        If the files containing the continuous seismograms do not have a sac 
        header, or if they have a different extension than .sac, the user can create 
        the two above files on his own, before calling the `prepare_data` method.
        (However, each file name should be in the format net.sta.loc.cha', where 
        net, sta, and loc (optional) are network, station, and location code, 
        respectively, and cha is channel (e.g., BHZ, BHN, or BHE). (If the component
        is R or T, the NE channels will be loaded and rotated during the calculation
        of the phase velocities.) For example, IV.AQU.00.BHZ.
        
        The stations.pickle file should contain a dictionary object where each key
        corresponds to a station code ($net.$sta) and each value should be a `tuple`
        containing latitude and longitude of the station. For example::
            
            { net1.sta1 : (lat1, lon1), net1.sta2 : (lat2, lon2)}
            
        The timespans.pickle file should contain a dictionary object where each key
        corresponds to a station code ($net.$sta) and each value should be a `tuple`
        containing the starttime and endtime (`obspy.UTCDateTime.timestamp`) of the 
        continuous seismogram associated with the station. For example::
            
            {net1.sta1 : (starttime1, endtime1), net1.sta2 : (starttime2, endtime2)}
    
    After the data have been "prepared", the receivers available can be 
    displayed by typing::
        
        an.plot_stations()
    
    Now we can calculate the dispersion curves automatically, provided that we
    pass a reference curve (i.e., `ndarray of shape (n, 2)`, where the 1st column 
    is frequency, the 2nd is phase velocity). For example::
        
        an.extract_dispcurves(refcurve, 
                              freqmin=0.01, 
                              freqmax=0.5, 
                              cmin=1.5,
                              cmax=4.5, 
                              distmax=2000, 
                              window_length=3600, 
                              overlap=0.5, 
                              min_no_days=30)
        
    The above will calculate the dispersion curves for all combinations of
    station pairs for which the inter-station distance is < 2000 km and at least
    30 days of simultaneous recordings are available. Cross-correlations will
    be computed in the frequency range 0.01-0.5 Hz on 50%-overlapping time
    windows. The results (`ndarrays of shape (m, 2)`, where the 1st column is
    frequency and the 2nd is phase velocity) will be saved to 
    /path/to/an_velocity/Z/dispcurves.
    """
    
    def __init__(self, src, savedir=None, component='Z', verbose=True):
        self.src = src
        savedir = os.path.dirname(src) if savedir is None else savedir
        self.verbose = verbose
        if component not in ['Z', 'R', 'T']:
            exception = '@component@ should be either `Z` (Rayleigh, vertical), \
                         `T` (Love, transverse), or `R` (Rayleigh, radial)'
            raise Exception(exception)
        self.component = component
        self.savedir = os.path.join(savedir, 'an_velocity', component)
        os.makedirs(self.savedir, exist_ok=True)
        self.files = self.get_files()
        
    
    def __repr__(self):
        return str(self)
    
    
    def __str__(self):
        phase = 'love'if self.component=='T' else 'rayleigh'
        if phase=='rayleigh' and self.component=='R':
            phase = phase + ' (radial)'
        string = '\n%s-WAVE VELOCITY FROM SEISMIC AMBIENT NOISE'%(phase.upper())
        separators = len(string)
        string += '\n%s'%('='*separators)
        stations = set(['.'.join(i.split('.')[:2]) for i in self.files])
        string += '\nRECEIVERS: %s'%(len(stations))
        try:
            done = len(os.listdir(os.path.join(self.savedir, 'dispcurves')))
        except FileNotFoundError:
            done = 0
        pairs = len(stations) * (len(stations)-1) // 2
        string += '\nSTATION PAIRS AVAILABLE: %s'%(pairs)
        string += '\nDISPERSION CURVES RETRIEVED: %s'%(done)
        string += '\n%s'%('='*separators)
        string += '\nSOURCE DIR: %s'%self.src
        string += '\nSAVE DIR: %s'%self.savedir
        return string
    
    
    def get_files(self):
        """ 
        Retrieves the files to be processed for extracting the phase velocities
        
        Returns
        -------
        files : list of str
            e.g. ['net1.sta1.00.BHZ.sac', 'net1.sta2.00.BHZ.sac']
        """
        files = []
        components = ['HZ'] if self.component=='Z' else ['HE', 'HN']
        for file in sorted(os.listdir(self.src)):
            channel = file.split('.')[-2][1:]
            if channel in components:
                files.append(file)
        return files
    
    
    def get_times_and_coords(self, files):
        """ 
        Retrieves the geographic coordinates and the starting and ending time
        associated with each continuous seismogram
        
        Parameters
        ----------
        files: list of str
            Names of the files corresponding with the continuous seismograms,
            located in the `src` directory
            
        Returns
        -------
        times : dict
            each key corresponds to a station code ($network_code.$station_code) 
            and each value is a tuple containing the starttime and endtime 
            (in timestamp format, see obspy documentation on UTCDateTime.timestamp) 
            of the continuous seismogram associated with the station. For
            example::
                
                { net1.sta1 : (starttime1, endtime1),
                  net1.sta2 : (starttime2, endtime2),
                  net2.sta3 : (starttime3, endtime3)
                  }
                
        coords : dict
            each key corresponds to a station code ($network_code.$station_code) 
            and each value is a tuple containing latitude and longitude of the 
            station. For example::
                
                { net1.sta1 : (lat1, lon1),
                  net1.sta2 : (lat2, lon2),
                  net2.sta3 : (lat3, lon3)
                  }
        """
        times = {}
        coords = {}
        for file in files:
            station_code = '.'.join(file.split('.')[:2])
            if station_code in coords:
                continue
            tr = _read(os.path.join(self.src, file), 
                       headonly=True,
                       verbose=self.verbose)[0]
            lat, lon = tr.stats.sac.stla, tr.stats.sac.stlo
            coords[station_code] = (lat, lon)
            starttime = tr.stats.starttime.timestamp
            endtime = tr.stats.endtime.timestamp
            times[station_code] = (starttime, endtime)
            if self.verbose:
                print(station_code, tr.stats.starttime, tr.stats.endtime)
        
        return times, coords

            
    def prepare_data(self, recompute=False):
        """ 
        Saves to disk the geographic coordinates and the starting and ending 
        time associated with each continuous seismogram. These are saved to
        $self.savedir/stations.pickle and $self.savedir/timespans.pickle,
        respectively. 
        
        The stations.pickle file contains a dictionary object where each key
        corresponds to a station code ($network_code.$station_code) and each 
        value is a tuple containing latitude and longitude of the station. 
        For example::
            
            { net1.sta1 : (lat1, lon1),
              net1.sta2 : (lat2, lon2),
              net2.sta3 : (lat3, lon3)
              }
            
        The timespans.pickle file contains a dictionary object similar to the
        above, where each value is a tuple containing the starttime and endtime 
        (`obspy.UTCDateTime.timestamp`) of the continuous seismogram associated 
        with the station. For example::
            
            { net1.sta1 : (starttime1, endtime1),
              net1.sta2 : (starttime2, endtime2),
              net2.sta3 : (starttime3, endtime3)
              }
        
        Parameters
        ----------
        recompute : bool
            If `True`, the station coordinates and times will be removed from
            disk and recalculated. Otherwise (default), if they are present,
            they will be loaded into memory, avoiding any computation. This
            parameter should be set to `True` whenever one has added new files
            to the source directory (and has not updated manually stations.pickle
            and timespans.pickle).

        Notes
        -----
        If `component` is 'Z', only continuous seismograms recorded on the 
        vertical component are considered (e.g., BHZ).

        If `component` is either 'R' or 'T', however, for a given station to 
        be considered in the calculation of phase velocities both the 
        horizontal components (N and E, e.g. BHN and BHE) should be 
        available. These will be rotated so as to analyse the transverse 
        (T, for Love) and radial (R, for radially-polarized Rayleigh waves) 
        components
        """
        savecoords = os.path.join(self.savedir, 'stations.pickle')
        savetimes = os.path.join(self.savedir, 'timespans.pickle')
        if recompute:
            remove_file(savecoords)
            remove_file(savetimes)
        if not os.path.exists(savecoords) or not os.path.exists(savetimes):
            times, coords = self.get_times_and_coords(self.files)
            save_pickle(savecoords, coords)
            save_pickle(savetimes, times)
        else:
            coords = load_pickle(savecoords)
            times = load_pickle(savetimes)
        self.stations = coords
        self.times = times
            
    
    @classmethod
    def convert_to_kms(cls, dispcurve):
        """ Converts from m/s to km/s the dispersion curve (if necessary).
        
        Parameters
        ----------
        dispcurve : ndarray of shape (n, 2)
            The 1st column (typically frequency or period) is ignored. The
            2nd should be velocity. If the 1st value of velocity
            divided by 10 is greater than 1, the 2nd column is divided by
            1000. Otherwise, the dispersion curve is left unchanged.
            
        Returns
        -------
        dispcurve : ndarray of shape (n, 2)
        """
        if dispcurve[0, 1] / 10 > 1:
            dispcurve = np.column_stack((dispcurve[:, 0], dispcurve[:, 1]/1000))
        return dispcurve
    
    
    def extract_dispcurves(self, refcurve, freqmin=0.01, freqmax=0.5, cmin=1.5,
                           cmax=4.5, distmax=None, window_length=3600, 
                           overlap=0.5, min_no_days=30, reverse_iteration=False,
                           save_xcorr=False, plotting=False, manual_picking=False,
                           show=True):
        """ 
        Automatic extraction of the dispersion curves for all available pairs
        of receivers.
        
        The results are saved to $self.savedir/dispcurves in .npy format,
        and consist of `ndarrays of shape (n, 2)`, where the 1st column is
        frequency and the 2nd phase velocity (in m/s).
        
        The routine iterates over all the available combinations of station
        pairs and, for each one, (i) computes the cross spectrum (in the 
        frequency domain) by ensamble averaging the cross correlations calculated 
        over relatively small (and possibly overlapping, see `overlap`) time 
        windows (see `window_length`), (ii) filters the cross-spectrum using 
        a "velocity" filter [1]_, and (iii) extracts a smooth dispersion 
        curve by comparison of the zero-crossings of the cross-spectrum with 
        those of the Bessel function associated with the station pair in 
        question [2]_.
        
        Parameters
        ----------
        refcurve : ndarray of shape (n, 2)
            Reference curve used to extract the dispersion curves. The 1st
            column should be frequency, the 2nd velocity (in either
            km/s or m/s). The reference curve is automatically converted to
            km/s, the physical units employed in the subsequent analysis
            
        freqmin, freqmax : float
            Minimum and maximum frequency analysed by the algorithm (default
            are 0.01 and 0.5 Hz). The resulting dispersion curves will be limited
            to this frequency range
            
        cmin, cmax : float
            Estimated velocity range (in km/s) spanned by the dispersion curves 
            (default values are 1.5 and 4.5)   
            
        distmax : float
            Maximum inter-station distance (in km), beyond which a given station 
            pair is not considered in the calculation of the phase velocities
            
        window_lenght : float
            Length of the time windows (in s) used to perform the cross 
            correlations
            
        overlap : float
            Should be >=0 and <1 (strictly smaller than 1). Rules the extent
            of overlap between one time window and the following [3]_. Default 
            is 0.5
            
        min_no_days : int, float
            Minimum number of simultaneous recordings available for a given
            station pair to be considered for the extraction of phase velocity.
            Default is 30
            
        reverse_iteration : bool
            If `True`, the list of combinations of station pairs will be 
            iterated over reversely. This can be useful to run two processes in
            parallel so as to halve the computation times (one function call
            setting `reverse_iteration`=`False`, the other, in a different process,
            setting it to `True`). Default is `False`     
        
        save_xcorr : bool
            If `True`, cross-correlations will be saved in $self.savedir/xcorr
        
        plotting : bool
            If `True`, a figure is created for each retrieved dispersion curve.
            This is automatically displayed and saved in $self.savedir/figures
            
        manual_picking : bool
            If `True`, the user is required to pick each dispersion curve manually.
            The picking is carried out through an interactive plot
            
        show : bool
            If `False`, the figure is closed before being shown. This is only
            relevant when the user has the interactive plot enabled: in that
            case, the figure will not be displayed. Default is `True`
            
        Returns
        -------
        None
            The dispersion curves are saved to $self.savedir/dispcurves

        References
        ----------
        .. [1] Magrini & Boschi, (2021), Surface‐Wave Attenuation From Seismic Ambient Noise: 
            Numerical Validation and Application, JGR 

        .. [2] Kästle et al. 2016, Two-receiver measurements of phase velocity: cross-
            validation of ambient-noise and earthquake-based observations, GJI

        .. [3] Seats et al. 2012, Improved ambient noise correlation functions using 
            Welch's method, GJI
        """
        def percentage_done(no_pairs, no_done):
            return '\nPERCENTAGE DONE: %.2f\n'%(no_done/no_pairs * 100)
        
        def load_done(file):
            if os.path.exists(file):
                return set([i.strip() for i in open(file, 'r')])
            else:
                return set()
    
        def update_done(sta1, sta2):
            with open(save_done, 'a') as f:
                f.write('%s__%s\n'%(sta1, sta2))
            # done.add('%s__%s'%(sta1, sta2))
                
        def station_pairs_generator(stations, reverse_iteration=False):
            sort = lambda iterable, reverse: sorted(iterable, reverse=reverse)
            station_pairs = (pair for pair in sort(it.combinations(sorted(stations), 2),
                                                   reverse_iteration))
            for pair in station_pairs:
                yield pair
        
        def read_stream(folder, file, component):
            st = _read(os.path.join(folder, file), verbose=self.verbose)
            if st is None:
                return
            if component=='R' or component=='T':
                cha = file[-5]
                new_cha = 'N' if cha=='E' else 'E'
                new_file = file.replace('%s.sac'%cha, '%s.sac'%new_cha)
                try:
                    st += _read(os.path.join(folder, new_file), verbose=self.verbose)
                except TypeError:
                    return
            return st
        
        def get_trace(st, azimuth, component):
            if component == 'Z':
                return st[0]
            st = rotate_stream(st, method='NE->RT', back_azimuth=(azimuth+180)%360)
            return st.select(component=component)[0]
        
        def dist_az_backaz(stations, sta1, sta2):
            stla1, stlo1 = stations[sta1]
            stla2, stlo2 = stations[sta2]
            dist, az, baz = gps2dist_azimuth(stla1, stlo1, stla2, stlo2)
            dist /= 1000.
            return dist, az, baz            
        
        def plot(dispcurve, refcurve, crossings, xcorr, sta1, sta2, dist, 
                 days_overlap, savefig=None, show=True):
            
            suptitle = '%s - %s | '%(sta1, sta2)
            suptitle += 'Dist: %.1f km | Overlapping days: %.1f'%(dist, days_overlap)
            plt.figure(figsize=plt.figaspect(0.4))
            plt.subplot(1, 2, 1)
            plt.plot(crossings[:,0], crossings[:,1], lw=0, marker='o', 
                     label='Zero crossings', markersize=5)
            plt.plot(refcurve[:,0], refcurve[:,1], 'k', lw=1, label='Reference')
            plt.plot(dispcurve[:,0], dispcurve[:,1], 'r', lw=1.5, label='Retrieved')
            plt.ylim(cmin, cmax)
            plt.ylabel('Phase velocity [km/s]')
            plt.xlim(freqmin, freqmax)
            plt.xlabel('Frequency [Hz]')
            plt.legend(loc='upper right', framealpha=0.9, handlelength=1)
            plt.grid(alpha=0.5)
            plt.title('Dispersion Curve')

            plt.subplot(1, 2, 2)
            plt.plot(xcorr[:,0], xcorr[:,1].real, 'k', lw=1.5, label='Real')
            plt.plot(xcorr[:,0], xcorr[:,1].imag, 'gray', lw=1.5, label='Imag')
            plt.xlim(freqmin, freqmax)
            plt.xlabel('Frequency [Hz]')
            plt.grid(alpha=0.5)
            plt.legend(loc='upper right', framealpha=0.9, handlelength=1)
            plt.title('Cross Spectrum')

            
            plt.tight_layout()
            plt.subplots_adjust(top=0.85)
            plt.suptitle(suptitle, y=0.98)
            if savefig is not None:
                plt.savefig(savefig, dpi=(200))
            if not show:
                plt.close()
            else:
                plt.show()
        
        save_pv = os.path.join(self.savedir, 'dispcurves')
        save_tmp = os.path.join(self.savedir, 'tmp')
        save_done = os.path.join(save_tmp, 'DONE.txt')
        save_corr = os.path.join(self.savedir, 'xcorr')
        save_fig = os.path.join(self.savedir, 'figures')
        os.makedirs(save_pv, exist_ok=True)
        os.makedirs(save_tmp, exist_ok=True)
        if save_xcorr:
            os.makedirs(save_corr, exist_ok=True)
        if plotting:
            os.makedirs(save_fig, exist_ok=True)
        refcurve = self.convert_to_kms(refcurve)
        
        files_dict = {'.'.join(i.split('.')[:2]):i for i in self.files}
        horizontal_polarization = False if self.component=='Z' else True
        npairs = int(len(self.stations) * (len(self.stations)-1) / 2)
        station_pairs = station_pairs_generator(self.stations, reverse_iteration)
        
        sta1_code = None
        done = load_done(save_done)
        for ndone, (sta1, sta2) in enumerate(station_pairs):
            if '%s__%s'%(sta1, sta2) in done:
                continue
            if not ndone % 100:
                done = load_done(save_done)
                if self.verbose:
                    print(percentage_done(npairs, len(done)))
                    
            sac1, sac2 = files_dict[sta1], files_dict[sta2]     
            (start1, end1), (start2, end2) = self.times[sta1], self.times[sta2]
            days_overlap = (min([end1, end2]) - max([start1, start2])) / 86400
            if days_overlap < min_no_days:
                update_done(sta1, sta2)
                continue
                        
            dist, az, baz = dist_az_backaz(self.stations, sta1, sta2)
            if distmax is not None and dist>=distmax:
                update_done(sta1, sta2)   
                continue
            
            if self.verbose:
                print(sta1, sta2)

            if sta1_code is None or sta1_code!=sta1:
                st1 = read_stream(folder=self.src, 
                                  file=sac1, 
                                  component=self.component)
                if st1 is None:
                    update_done(sta1, sta2)
                    continue
                sta1_code = sta1
            st2 = read_stream(folder=self.src, 
                              file=sac2, 
                              component=self.component)
            if st2 is None:
                update_done(sta1, sta2)
                continue
            tr1 = get_trace(st1, azimuth=az, component=self.component)
            tr2 = get_trace(st2, azimuth=az, component=self.component)      
            dispcurve_file = os.path.join(save_pv, '%s__%s.npy'%(tr1.id, tr2.id))
            if os.path.exists(dispcurve_file):
                update_done(sta1, sta2)
                continue
            
            xcorr_file = os.path.join(save_corr, '%s__%s.npy'%(tr1.id, tr2.id))
            fig_file = os.path.join(save_fig, '%s__%s.png'%(tr1.id, tr2.id))
        
            try:
                freq, xcorr = noisecorr(tr1, tr2, window_length=window_length, 
                                        overlap=overlap)
                xcorr_smooth = velocity_filter(freq, xcorr, dist, cmin=cmin, cmax=cmax)
                crossings, phase_vel = extract_dispcurve(
                        freq, 
                        xcorr_smooth, 
                        dist, 
                        refcurve,
                        freqmin=freqmin, 
                        freqmax=freqmax, 
                        cmin=cmin, 
                        cmax=cmax,
                        horizontal_polarization=horizontal_polarization,
                        manual_picking=manual_picking,
                        sta1=sta1,
                        sta2=sta2
                        )
                if save_xcorr:
                    np.save(xcorr_file, np.column_stack((freq, xcorr)))
            except:
                update_done(sta1, sta2)
                continue
                        
            if plotting:
                plot(dispcurve=phase_vel, 
                     refcurve=refcurve, 
                     crossings=crossings, 
                     xcorr=np.column_stack((freq, xcorr_smooth)), 
                     sta1=tr1.id, 
                     sta2=tr2.id, 
                     dist=dist, 
                     days_overlap=days_overlap, 
                     savefig=fig_file,
                     show=show)
                
            phase_vel[:,1] *= 1000 # converts to m/s
            np.save(dispcurve_file, phase_vel)
            update_done(sta1, sta2)
            
    
    def prepare_input_tomography(self, savedir, period, min_no_wavelengths=2,
                                 outfile='input_%.2fs.txt'):
        """ 
        Prepares a .txt file for each specified period, to be used for 
        calculating phase-velocity maps using 
        :class:`seislib.tomography.tomography.SeismicTomography`.
        
        Parameters
        ----------
        savedir : str
            Absolute path to the directory where the file(s) is (are) saved.
            If `savedir` does not exist, it will be created
            
        period : float, iterable of floats
            Period (or periods) at which the dispersion curves will be 
            interpolated (see :meth:`interpolate_dispcurves`)
            
        min_no_wavelengths : float
            Ratio between the estimated wavelength :math:`\lambda = T c_{ref}`
            of the surface-wave at a given period *T* and interstation distance
            :math:`\Delta`. Whenever this ratio is > `min_no_wavelength`, 
            the period in question is not used to retrieve a dispersion measurement. 
            Values < 1 are suggested against. Default is 2
            
        outfile : str
            Format for the file names. It must include either %s or %.Xf (where
            X is integer), since this will be replaced by each period analysed
            (one for file)

        Examples
        --------
        Assume you calculated dispersion curves from your data, which are 
        located in /path/to/data, and you had initialized your 
        class:`AmbientNoiseVelocity` instance, to calculate inter-station 
        dispersion curves, as::

            an = AmbientNoiseVelocity(src=SRC,
                                      component='Z')
            an.prepare_data()

        You can, even if the computation of the dispersion curves is not 
        finished, use the above instance to extract the information needed
        to produce phase-velocity maps as follows::

            savedir = /arbitrary/path/to/savedir
            periods = [5, 6, 8, 10, 12, 15, 20]
            an.prepare_input_tomography(savedir=savedir,
                                        period=periods)

        The above will save one txt file for each specified period.
        """
        os.makedirs(savedir, exist_ok=True)
        period = np.array([period]) if np.isscalar(period) else np.array(period)
        codes, coords, velocity = self.interpolate_dispcurves(1/period)
        dist = SeismicTomography.gc_distance(*coords.T)
        wavelength = velocity * period
        ratios = dist.reshape(-1, 1) / wavelength
        velocity_final = np.where(ratios>min_no_wavelengths, velocity, np.nan)

        for iperiod, p in enumerate(period):
            save = os.path.join(savedir, outfile%p)
            vel = velocity_final[:, iperiod]
            notnan = np.flatnonzero(~np.isnan(vel))
            if self.verbose:
                print('Measurements at %.2fs:'%p, notnan.size)
            np.savetxt(save, np.column_stack((coords[notnan], vel[notnan])))
        
    
    def interpolate_dispcurves(self, frequency):
        """ 
        Interpolates the dispersion curves found at $self.savedir/dispcurves
        at the specified frequency. (No extrapolation is made.)
        
        Parameters
        ----------
        frequency : float, iterable of floats
            Frequency (or frequencies) at which the dispersion curves will be 
            interpolated
            
        Returns
        -------
        codes : ndarray of shape (n, 2)
            Codes associated with the station pairs for which a dispersion curve
            has been calculated
            
        coords : ndarray of shape (n, 4)
            Coordinates (lat1, lon1, lat2, lon2) of the station pairs corresponding
            to the station codes
            
        measurements : ndarray of shape (n, f)
            Phase velocities calculated for each station pair in `coords` at
            the specified `frequency`. f is the number of frequencies

            .. note::
                *measurements* could contain nans

        Examples
        --------
        Assume you calculated dispersion curves from your data, which are 
        located in /path/to/data, and you had initialized your 
        class:`AmbientNoiseVelocity` instance, to calculate inter-station 
        dispersion curves, as::

            an = AmbientNoiseVelocity(src=SRC,
                                      component='Z')
            an.prepare_data()

        Even if the computation of the dispersion curves is not yet 
        finished, you can use the above instance to interpolate all the
        available dispersion curves as follows::

            frequency = [1/30, 1/20, 1/10, 1/5]
            codes, coords, measurements = an.interpolate_dispcurves(frequency)        
        """
        def display_progress(no_files, done, verbose=False):
            if verbose and not done % int(0.05*no_files + 1):
                print('FILES PROCESSED: %d/%d'%(done, no_files))
                    
        src = os.path.join(self.savedir, 'dispcurves')
        files = [i for i in sorted(os.listdir(src)) if i.endswith('npy')]
        files_size = len(files)
        freq_size = 1 if np.isscalar(frequency) else len(frequency)
        measurements = np.zeros((files_size, freq_size))        
        coords = np.zeros((files_size, 4))
        codes = []
        for i, file in enumerate(files):
            display_progress(no_files=files_size, done=i, verbose=self.verbose)
            freq, vel = np.load(os.path.join(src, file)).T
            interp_vel = interp1d(freq, vel, bounds_error=False)(frequency)
            measurements[i] = interp_vel
            code1, code2 = file.split('.npy')[0].split('__')
            sta1 = '.'.join(code1.split('.')[:2])
            sta2 = '.'.join(code2.split('.')[:2])
            coords[i] = (*self.stations[sta1], *self.stations[sta2])
            codes.append((sta1, sta2))
        return np.array(codes), coords, measurements
    
    
    def plot_stations(self, ax=None, show=True, oceans_color='water', 
                      lands_color='land', edgecolor='k', projection='Mercator',
                      resolution='110m', color_by_network=True, legend_dict={}, 
                      **kwargs):
        """ Maps the seismic receivers for which data are available
        
        Parameters
        ----------
        ax : cartopy.mpl.geoaxes.GeoAxesSubplot
            If not `None`, the receivers are plotted on the `GeoAxesSubplot` instance. 
            Otherwise, a new figure and `GeoAxesSubplot` instance is created
            
        show : bool
            If `True`, the plot is shown. Otherwise, a `GeoAxesSubplot` instance is
            returned. Default is `True`
            
        oceans_color, lands_color : str
            Color of oceans and lands. The arguments are ignored if ax is not
            `None`. Otherwise, they are passed to `cartopy.feature.NaturalEarthFeature` 
            (to the argument 'facecolor'). Defaults are 'water' and 'land'
            
        edgecolor : str
            Color of the boundaries between, e.g., lakes and land. The argument 
            is ignored if ax is not `None`. Otherwise, it is passed to 
            `cartopy.feature.NaturalEarthFeature` (to the argument 'edgecolor'). 
            Default is 'k' (black)
            
        projection : str
            Name of the geographic projection used to create the `GeoAxesSubplot`.
            (Visit the `cartopy website 
            <https://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html>`_ 
            for a list of valid projection names.) If ax is not None, `projection` 
            is ignored. Default is 'Mercator'
            
        resolution : {'10m', '50m', '110m'}
            Resolution of the Earth features displayed in the figure. Passed to
            `cartopy.feature.NaturalEarthFeature`. Valid arguments are '110m',
            '50m', '10m'. Default is '110m'. Ignored if `ax` is not `None`.
            
        color_by_network : bool
            If `True`, each seismic network will have a different color in the
            resulting map, and a legend will be displayed. Otherwise, all
            stations will have the same color. Default is `True`
        
        legend_dict : dict, optional
            Dictionary of keyword arguments passed to `matplotlib.pyplot.legend`
        
        **kwargs : dict, optional
            Additional keyword arguments passed to `matplotlib.pyplot.scatter` 
            
        Returns
        -------
        If `show` is True, None, else `ax`, i.e. the GeoAxesSubplot
        """        
        return plot_stations(stations=self.stations,
                             ax=ax, 
                             show=show, 
                             oceans_color=oceans_color, 
                             lands_color=lands_color, 
                             edgecolor=edgecolor, 
                             projection=projection,
                             resolution=resolution,
                             color_by_network=color_by_network, 
                             legend_dict=legend_dict,
                             **kwargs)      






#cython: language_level=3
"""
@author: Fabrizio Magrini
@email1: fabrizio.magrini@uniroma3.it
@email2: fabrizio.magrini90@gmail.com


The module assumes a unit Earth
"""


cimport cython
from cython cimport cdivision, boundscheck, wraparound
from libc.math cimport pi as PI
from libc.math cimport sqrt



@cdivision
cdef double radians(double deg):
    return deg * PI / 180


@cdivision
cdef double degrees(double rad):
    return rad * 180 / PI


@boundscheck(False)
@wraparound(False)
cdef void cross_product_3d(double[3] v1, double[3] v2, double[3] result):
    result[0] = v1[1]*v2[2] - v1[2]*v2[1]
    result[1] = v1[2]*v2[0] - v1[0]*v2[2]
    result[2] = v1[0]*v2[1] - v1[1]*v2[0]


@boundscheck(False)
@wraparound(False)
cdef double norm_3d(double[3] v):
    return sqrt(v[0]**2 + v[1]**2 + v[2]**2)




#cython: language_level=3
"""
@author: Fabrizio Magrini
@email1: fabrizio.magrini@uniroma3.it
@email2: fabrizio.magrini90@gmail.com


The module assumes a unit Earth
"""


cimport cython
from cython cimport cdivision, boundscheck, wraparound, binding
from libc.math cimport cos, sin, tan, acos, asin, atan, atan2
from libc.math cimport pi as PI
from libc.math cimport sqrt, fabs
from libcpp cimport bool as bool_cpp
from libcpp.list cimport list as cpplist
import numpy as np
cimport numpy as np
from seislib.tomography._ray_theory._math cimport radians, degrees
from seislib.tomography._ray_theory._math cimport cross_product_3d, norm_3d
cdef double TWOPI = 2 * PI


@boundscheck(False)
@wraparound(False)
cdef void cartesian_coordinates(double[2] latlon, double[3] result):
    """
    [lat, lon] in radians
    """
    result[0] = cos(latlon[0]) * cos(latlon[1])
    result[1] = cos(latlon[0]) * sin(latlon[1])
    result[2] = sin(latlon[0])


@boundscheck(False)
@wraparound(False)
cdef void spherical_coordinates(double[3] xyz, double[2] result):
    """
    cartesian coordinates: [x, y, z]
    """
    result[0] = asin(xyz[2])
    result[1] = atan2(xyz[1], xyz[0])


@cdivision(True)
@boundscheck(False)
@wraparound(False)
cdef void gc_arc_midpoint(double[2] latlon_1, double[2] latlon_2, double[2] midpoint):
    """
    v1, v2 : cartesian coordinates of the two points on the great circle
    """
    cdef double norm
    cdef double[3] xyz_1, xyz_2, xyz_midpoint
    cartesian_coordinates(latlon_1, xyz_1)
    cartesian_coordinates(latlon_2, xyz_2)
    xyz_midpoint[:] = [ xyz_1[0]+xyz_2[0], xyz_1[1]+xyz_2[1], xyz_1[2]+xyz_2[2] ]
    norm = norm_3d(xyz_midpoint)
    xyz_midpoint[:] = [ xyz_midpoint[0]/norm, xyz_midpoint[1]/norm, xyz_midpoint[2]/norm ]
    spherical_coordinates(xyz_midpoint, midpoint)
    

@boundscheck(False)
@wraparound(False)
cdef void great_circle_plane(double[2] r1, double[2] r2, double[3] params):
    """
    Plane through r1, r2 (lat, lon in radians), and origin sphere
    """
    params[0] = cos(r1[0])*sin(r2[0])*sin(r1[1]) - cos(r2[0])*sin(r1[0])*sin(r2[1])
    params[1] = cos(r2[0])*sin(r1[0])*cos(r2[1]) - cos(r1[0])*sin(r2[0])*cos(r1[1])
    params[2] = cos(r1[0]) * cos(r2[0]) * sin(r2[1]-r1[1])


@cdivision(True)
@boundscheck(False)
@wraparound(False)
cdef double great_circle_distance(double[2] latlon_1, double[2] latlon_2):
    """
    Calculate the great circle distance between two points 
    on the earth.
    [lat1, lon1], [lat2, lon2] : radians
    """
    cdef double lat1 = latlon_1[0]
    cdef double lon1 = latlon_1[1]
    cdef double lat2 = latlon_2[0]
    cdef double lon2 = latlon_2[1]
    cdef double dlon = lon1 - lon2
    cdef double dlat = lat1 - lat2
    return 2*asin(sqrt((sin((dlat)/2))**2 + cos(lat1)*cos(lat2)*(sin((dlon)/2))**2))
    

@cdivision(True)
@boundscheck(False)
@wraparound(False)
cdef double gc_meridian_intersection(double[3] gc_plane, double parallel1, 
                                     double parallel2, double meridian):
    cdef double a_12 = gc_plane[0]
    cdef double b_12 = gc_plane[1]
    cdef double c_12 = gc_plane[2]               
    lat_intersect = -atan( a_12*cos(meridian)/c_12 + b_12*sin(meridian)/c_12 )
    if parallel1 <= lat_intersect <= parallel2: 
        return lat_intersect 
    return 9999


@cdivision(True)
@boundscheck(False)
@wraparound(False) 
cdef void gc_parallel_intersection(double[2] r1, double[2] r2, double meridian1, 
                                   double meridian2, double parallel, double[2] lon_intersections):
    lon_intersections[:] = [9999, 9999]
    cdef double lat1 = r1[0]
    cdef double lon1 = r1[1]
    cdef double lat2 = r2[0]
    cdef double lon2 = r2[1]
    cdef double l12 = lon1 - lon2
    cdef double a, b, c, ab_norm
    cdef double lon, lon_intersection1, lon_intersection2
    a = sin(lat1)*cos(lat2)*cos(parallel)*sin(l12)
    b = sin(lat1)*cos(lat2)*cos(parallel)*cos(l12) - cos(lat1)*sin(lat2)*cos(parallel)
    c = cos(lat1)*cos(lat2)*sin(parallel)*sin(l12)
    lon = atan2(b, a)
    ab_norm = sqrt(a**2 + b**2)
    if (fabs(c) > ab_norm):
        return
    dlon = acos(c / ab_norm)
    lon_intersection1 = mod(lon1+dlon+lon+PI, TWOPI) - PI
    lon_intersection2 = mod(lon1-dlon+lon+PI, TWOPI) - PI
    if meridian1 <= lon_intersection1 <= meridian2:
        lon_intersections[0] = lon_intersection1
    if meridian1 <= lon_intersection2 <= meridian2:
        lon_intersections[1] = lon_intersection2


cpdef double mod(double a, double b):
    return a % b






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
=======
SeisLib
=======
"""

from __future__ import print_function
import sys
import os
from .__version__ import __version__
from . import an
from . import colormaps
from . import eq
from . import exceptions
from . import plotting
from . import tomography
from . import utils






"""
===========================================
Plotting (:mod:`seislib.plotting`)
===========================================

SeisLib provides support for plotting the results obtained from
each of its modules, through the functions contained in 
:mod:`seislib.plotting`. :mod:`seislib.plotting` relies on
`CartoPy <https://scitools.org.uk/cartopy/docs/latest/>`_,
and allows one to

- Display geographic location of seismic receivers and epicenters

- Plot inter-station (or epicenter-station) great-circle paths
  for which velocity measurements are available. These can also 
  be colored according to the measured velocity

- Display spatial variations in any physical parameter defined on
  unstructured blocks (e.g., equal-area grids)

- Display Earth's features (e.g., coastlines) in a very simple (high-level) 
  fashion
    
- Easily add inset axes and colorbars

- Combine the above features to obtain high-quality figures.

"""
from ._plotting import *

__all__ = [
    'add_earth_features',
    'add_earth_features_GSHHS',
    'add_inset_axis',
    'colormesh',
    'contour',
    'contourf',
    'make_colorbar',
    'plot_colored_rays',
    'plot_events',
    'plot_map',
    'plot_rays',
    'plot_stations',
    'scientific_label',    
    ]




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Download of Teleseismic-Earthquake Recordings
============================================= 

The below class allows for automated downloads of seismograms
recording teleseismic surface waves. The data will be saved on disk
in the format required by :class:`seislib.eq.eq_velocity.EQVelocity` and
:class:`seislib.eq.eq_velocity.TwoStationMethod` to calculate inter-station
dispersion curves based on a two-station method.

"""

import os
import warnings
import time
import socket
from obspy import read_inventory, Catalog
from obspy import UTCDateTime as UTC
from obspy.core import AttribDict
from obspy.geodetics import gps2dist_azimuth, degrees2kilometers
from obspy.clients.fdsn import Client
from obspy.clients.fdsn.header import FDSNNoDataException, FDSNException
from obspy.io.mseed import InternalMSEEDError
from http.client import IncompleteRead
from collections import defaultdict
from seislib.utils import resample, rotate_stream, remove_file
from seislib.plotting import plot_stations
warnings.simplefilter(action='ignore')
ONE_YEAR = 365 * 86400



class EQDownloader:
    """ Downloads surface-wave data to be used via the two-station method

    Parameters
    ----------
    savedir : str
        Absolute path to the directory where the data will be saved. Inside 
        this directory, a folder named 'data' will be created. Inside 'data'
        the seismograms will be organized in subdirectories (one for each 
        seismic event) in the format net.sta.loc.cha.sac.
        
    inventory_name : str
        Name of the inventory (`obspy.core.inventory.inventory.Inventory`) 
        associated with the downloads. This will be saved in xml format
        in the `savedir`
        
    inv_provider : str
        Provider of station metadata, passed to `obspy.clients.fdsn.Client`. 
        Default is `iris`
        
    user, password : str, optional
        User name and password for accessing restricted data. Passed to 
        `obspy.clients.fdsn.Client`
        
    ev_provider : str
        Provider of earthquake metadata, passed to `obspy.clients.fdsn.Client`. 
        Default is `iris`
        
    vmin, vmax : float
        Minimum and maximum velocity of the surface waves. These values are
        used to establish the starting and ending time of the seismograms
        to be downloaded with respect to the origin time of given earthquake.
        Relatively loose limits are to be preferred. Default values are 1.5 
        and 5.5 (km/s), respectively
        
    sampling_rate : int, float
        Final sampling rate of the waveforms, in Hz. Default is 1 Hz
        
    units : {'DISP', 'VEL', 'ACC'}
        Physical units of the final waveforms. Can be either 'DISP', 'VEL',
        or 'ACC' (case insensitive). Default is 'DISP'.
        
    prefilter : (4, ) tuple of floats
        Bandapass filter applied to the waveforms before removing the
        intrument response. Default is (0.001, 0.005, 0.05, 0.4)
        
    attach_response : bool
        If `True`, the details of the instrument response are attached to the 
        header of the waveforms during their downloads. It can slow down the 
        downloads, but it will make errors due to response information less 
        likely. Default is `False`
        
    stations_config : dict
        Python dictionary containing optional keys and values passed to
        `obspy.clients.fdsn.Client.get_stations`.
        
    events_config : dict
        Python dictionary containing optional keys and values passed to
        `obspy.clients.fdsn.Client.get_events`. Default values are::
            distmin = 2223.9 (km, corresponding to 20°)
            distmax = 15567.3 (km, corresponding to 140°)
            depthmax = 50 (km)
            magmin = 6
            magmax = 8.5
            starttime = UTC(2000, 1, 1)
            endtime = UTC(2021, 1, 1)

        .. warning:: 
            The obspy key words 'minmag', 'maxmag', and 'maxdepth' have been 
            renamed as 'magmin', 'magmax', and 'depthmax', for better clarity 
            in the code. Passing to `events_config` one of the (obspy) key 
            words among 'minmag', 'maxmag', and 'maxdepth' will result in an 
            error.
        
        .. note::
            The above listed default values include 'distmin' and 'distmax' (in km). These refer 
            to the minimum and maximum distance of the epicenter from a given receiver. Note 
            that these two key words correspond to the obspy's 
            'minradius' and 'maxradius', with the only difference that the latters 
            are expressed in degrees. 'distmin' and 'distmax' have been introduced only for 
            achieving a higher consistency in the use of physical units throughout the code. 
            The user should be aware that, although the use of obspy's 'minradius' and 
            'maxradius' will not result in any error, their use is suggested against. 
            In fact, if 'distmin' and 'distmax' are not specified, their default value 
            will be used in the downloads. And if this default is more "restrictive" than 
            'minradius' and 'maxradius', 'minradius' and 'maxradius' will simpy be ignored 
            in the downloads.
        
    sleep_time_after_exception : float
        Time to wait (in s) after an `obspy.clients.fdsn.header.FDSNException`.
        The excpetion could be due to temporal issues of the client, so
        waiting a little bit before the next download could be useful to
        get things back to normality.
        
    verbose : bool
        If `True`, information on the downloads will be printed in console


    Attributes
    ----------
    savedir : str
        Absolute path of the directory that will be created
        
    staclient : obspy.clients.fdsn.Client
        Provider of station metadata and waveforms
        
    evclient : obspy.clients.fdsn.Client
        Provider of events metadata
        
    channels : list
        List of candidate channels for the downloads
        
    components : {3, 1}
        Either 3 (ZNE) or 1 (Z) 
        
    vmin, vmax : float
        Minimum and maximum expected velocity of the surface waves
        
    sampling_rate : float
        Sampling rate of the final seismograms
        
    prefilter : tuple
        Filter applied to the waveform before removal of the instrument response
        
    units : str
        Physical units of the final seismogram
        
    attach_response : bool     
        Whether or not the response information is attached to the waveforms
        during the download
        
    verbose : bool
        Whether or not progress information are displayed in the console
        
    sleep_time_after_exception : float
        The downloads are stopped for the specified time (in s) after a FDSN 
        exception
        
    distmin, distmax : float
        Minimum and maximum distance of a station from the epicenter (km)
        
    depthmax : float
        Maximum depth of the earthquakes (km)
        
    magmin, magmax : float
        Minimum and maximum magnitude
        
    starttime, endtime : obspy.core.utcdatetime.UTCDateTime
        Starttime and entime of the events catalogue
        
    stations_config, events_config : dict
        Dictionary object with information relevant to the download of the
        stations and events metadata
        
    inventory : obspy.core.inventory.inventory.Inventory
        Stations metadata
        
    _exceptions : collections.defaultdict(int)
        Dictionary object with information on the occurred exceptions
        
    _downloaded : int
        Number of waveforms downloaded
        
    _no_stations : int
        Number of stations to download
        
    _events_done : int
        Number of events for which the downloads have been finished


    Examples
    --------
    The following will download seismic data from all the seismic networks 
    and stations managed by iris within the region specified by minlatitude, 
    maxlatitude, minlongitude, maxlongitude. All such parameters should be passed to
    `stations_config`::

        stations_config=dict(
                channel='LH*,BH*,HH*',
                includerestricted=False,
                maxlatitude=12,
                minlatitude=-18,
                minlongitude=90,
                maxlongitude=140)    
            
    The above also specifies the preference of downloading 3-component seismograms (ZNE), 
    as implied by the asterisk sign. For a given receiver recording an earthquake, either 
    of the LH, BH, or HH channels will be downloaded, with priority given from left to right, 
    i.e., LH>BH>HH: if LH is not available, BH will be downloaded; if BH is not available, HH will 
    be downloaded.
        
    In this example, we use all earthquakes characterized by a 
    magnitude between 6 and 8.5, by a maximum depth of 50km, and generated 
    between the 1.1.2000 and 1.1.2021. The distance between each earthquake
    and a given receiver should be more than 20° (expressed in km) and less
    than 140° (in km)::

        from obspy import UTCDateTime as UTC
        from obspy.geodetics import degrees2kilometers

        events_config=dict(
                starttime=UTC(2000, 1, 1),
                endtime=UTC(2021, 1, 1),
                depthmax=50,
                magmin=6, 
                magmax=8.5,
                distmin=degrees2kilometers(20),
                distmax=degrees2kilometers(140),
                    )      
        
    We initialize the EQDownloader instance, and then start it::
        
        from seislib.eq import EQDownloader
        
        downloader = EQDownloader(savedir='/path/to/directory',
                                  inv_provider='iris',
                                  ev_provider='iris',
                                  inventory_name='iris.xml',
                                  sampling_rate=1,
                                  prefilter=(0.001, 0.005, 0.1, 0.4),
                                  vmin=1.5,
                                  vmax=5.5,
                                  units='disp',
                                  attach_response=False,
                                  stations_config=stations_config,
                                  events_config=events_config,
                                  verbose=True)        
        downloader.start()
    """


    def __init__(self, savedir, inventory_name, inv_provider='iris', user=None, 
                 password=None, ev_provider='iris', vmin=1.5, vmax=5.5, 
                 sampling_rate=1, units='disp', prefilter=(0.001, 0.005, 0.05, 0.4), 
                 attach_response=False, stations_config={}, events_config={},
                 sleep_time_after_exception=30, verbose=True):

        self.savedir = savedir
        os.makedirs(savedir, exist_ok=True)
        self.staclient = Client(inv_provider, user=user, password=password)
        self.evclient = Client(ev_provider)

        self.channels = stations_config.get('channel', None).split(',')
        self.components = 3 if self.channels[0][2]=='*' else 1
        self.sampling_rate = sampling_rate
        #self.npts = self.sampling_rate * self.duration
        self.vmin = vmin
        self.vmax = vmax
        self.prefilter = prefilter
        self.units = units
        self.attach_response = attach_response
            
        self.verbose = verbose
        self.sleep_time_after_exception = sleep_time_after_exception
        
        self.distmin = events_config.get('distmin', degrees2kilometers(20))
        self.distmax = events_config.get('distmax', degrees2kilometers(140))
        self.depthmax = events_config.get('depthmax', 50)
        self.magmin = events_config.get('magmin', 6)
        self.magmax = events_config.get('magmax', 8.5)
        self.starttime = events_config.get('starttime', UTC(2000, 1, 1))
        self.endtime = events_config.get('endtime', UTC(2021, 1, 1))
        keys = ['distmin', 'distmax', 'depthmax', 'magmin', 'magmax',
                'starttime', 'endtime']
        self.events_config = {k: v for k, v in events_config.items() \
                              if k not in keys}

        self.stations_config = stations_config
        inventory_path = os.path.join(savedir, inventory_name)
        if not os.path.exists(inventory_path):
            self.inventory = self.build_inventory(**stations_config)
            self.inventory.write(inventory_path, format="STATIONXML")
        else:
            self.inventory = read_inventory(inventory_path)
        
        self._exceptions = defaultdict(int)
        self._downloaded = 0
        self._events_done = 0
    
    
    def __str__(self):
        string = '\nWAVEFORMS DOWNLOADED: %d\n'%self._downloaded
        if len(self._exceptions) > 0:
            string += 'EXCEPTIONS:\n'
            for key, value in sorted(self._exceptions.items()):
                string += ''.join([' '*5, '%s: %d\n'%(key.upper(), value)])
        string += ''.join(['-'*40, '\n', ' '*5, '* '*3])
        string += 'EVENTS DONE: %d'%(self._events_done)
        string += ''.join([' *'*3, '\n'])
        string += ''.join(['-'*40, '\n'])
        return string

    def build_inventory(self, **kwargs):
        """ Builds an obspy inventory containing stations information
        
        The strategy is to first download inventory information at station level.
        Then, for each station, the instrument response is downloaded for each
        channel. This may increase the downloading time, but prevents possible
        "timed out errors".
        
        Parameters
        ----------
        **kwargs : dict, optional
            Additional key word arguments passed to the get_stations method of 
            `obspy.clients.fdsn.client`
            
        Returns
        -------
        inv : obspy.core.inventory.inventory.Inventory
        """
        def attach_channels(inv, **kwargs):
            
            tmp_kwargs = {i:j for i, j in kwargs.items() \
                          if i not in ['network', 'station']}
            for network in inv:
                for station in network:
                    if station.__dict__.get('creation_date') is None:
                        station.creation_date = station.start_date
                    while 1:
                        try:
                            tmp = self.staclient.get_stations(network=network.code, 
                                                              station=station.code, 
                                                              level='response',
                                                              **tmp_kwargs)
                            station.channels.extend(tmp[0][0].channels)
                            break
                        except FDSNNoDataException:
                            break
                        except FDSNException:
                            time.sleep(0.5)
                            continue
                        except TypeError:
                            continue
                        except ConnectionResetError:
                            time.sleep(0.5)
                            continue
                        except socket.timeout:
                            time.sleep(0.5)
                            continue
            return inv

        if self.verbose > 0:
            print('\n***COLLECTING STATIONS INFO***\n')
        inv = self.staclient.get_stations(**kwargs)
        if self.verbose:
            print(inv)
        return attach_channels(inv, **kwargs)

    @classmethod
    def inventory_iterator(cls, inventory, reverse=False):
        """ Generator function to iterate over an obspy inventory
        
        Parameters
        ----------
        inventory : obspy.core.inventory.inventory.Inventory  
        
        reverse : bool
            If `True`, the inventory will be iterated over from bottom to top
            
            
        Yields
        ------
        (2,) tuple 
            Yields network and station information at each iteration
        """
        def func(iterable, reverse=False):
            if reverse:
                return reversed(iterable)
            return iterable
        
        for network in func(inventory, reverse=reverse):
            for station in func(network, reverse=reverse):
                yield network, station
                
                
    @classmethod
    def event_coordinates_and_time(cls, event):
        """ Fetch event coordinates (lat, lon, depth) and origin time.
        
        Parameters
        ----------
        event : obspy.core.event.event.Event
        
        Returns
        -------
        (2,) Tuple
            (latitude, longitude, depth) and origin time in 
            `obspy.UTCDateTime.timestamp` format
    
        """
        otime = event.origins[0].time
        evla = event.origins[0].latitude
        evlo = event.origins[0].longitude
        evdp = event.origins[0].depth
        return (evla, evlo, evdp), otime
    
    
    @classmethod
    def station_coordinates(cls, station_info):
        """ Fetch station coordinates (lat, lon, elevation).
        
        Parameters
        ----------
        station_info : obspy.core.inventory.station.Station
        
        
        Returns
        -------
        (3,) Tuple of floats
            latitude, longitude, elevation
        """
        stla = station_info.latitude
        stlo = station_info.longitude
        stel = station_info.elevation
        return (stla, stlo, stel)


    @classmethod
    def get_event_info(cls, event):
        """ Fetch event information
        
        Parameters
        ----------
        event : obspy.core.event.event.Event
        
        Returns
        -------
        (2, ) Tuple
            (origin time, magnitude), (latitude, longitude, depth)
        """
        (evla, evlo, evdp), otime = cls.event_coordinates_and_time(event)
        mag = event.magnitudes[0].mag
        return (otime, mag), (evla, evlo, evdp)    
    
    
    def fetch_catalog(self, t1, t2, **kwargs):
        """ Fetch catalog of seismic events
        
        Parameters
        ----------
        t1, t2 : obspy.core.utcdatetime.UTCDateTime
            Starttime and endtime of the catalog
            
        **kwargs : dict, optional
            Additional key-word arguments passed to 
            `obspy.clients.fdsn.client.Client.get_events`
        
        Returns
        -------
        obspy.core.event.catalog.Catalog
            Catalog of seismic events
        """
        catalog = self.evclient.get_events(
                starttime=t1, 
                endtime=t2, 
                maxdepth=self.depthmax,
                minmagnitude=self.magmin,
                maxmagnitude=self.magmax,
                **kwargs
                )
        return Catalog(sorted(catalog, key=lambda ev: ev.origins[0].time))
      
        
    @classmethod     
    def station_was_active(cls, station, time):
        """ Wheater or not the seismic station was active at the given time
        
        Parameters
        ----------
        station : obspy.core.inventory.station.Station
        
        time : obspy.core.utcdatetime.UTCDateTime
        
        
        Returns
        -------
        bool
            Whether or not the receiver was active at the specified time
        """
        if station.start_date and (time<station.start_date):
            return False
        if station.end_date and (time>station.end_date):
            return False
        return True
    
    
    def active_channels(self, station):
        """ Channels available for the given station among those to download
        
        Parameters
        ----------
        station : obspy.core.inventory.station.Station
        
        Returns
        -------
        List
        """
        channels = set([channel.code[:2] for channel in station.channels])
        return [cha for cha in self.channels if cha[:2] in channels]
    
    
    def collect_waveforms(self, network, station, channels, starttime, endtime):
        """ Downloads obspy stream
        
        Parameters
        ----------
        network, station : str 
            network and station codes
            
        channels : iterable of str
            iterable containing the channels codes suited to the download. Each
            channel will be tried to be used in the downloads. The first successfull
            attempt determines the returned waveforms. (Higher priority is given to
            decreasing indexes)
            
        starttime, endtime : obspy.core.utcdatetime.UTCDateTime
            start and end time of the stream.
        
        Returns
        -------
        obspy.core.stream.Stream if the download is successful, else None
        """
        
        def gaps_are_present(st):
            return False if not st.get_gaps() else True
        
        for cha in channels:
            try:
                st = self.staclient.get_waveforms(network=network, 
                                                  station=station, 
                                                  location='*', 
                                                  channel=cha, 
                                                  starttime=starttime, 
                                                  endtime=endtime,
                                                  attach_response=self.attach_response)
                if gaps_are_present(st):
                    self._exceptions['Gaps in stream'] += 1
                    return None
#                st.merge(method=1, interpolation_samples=2, fill_value=0)
                return st               
                
            except FDSNNoDataException:
                self._exceptions['FDSNNoDataException'] += 1
            except InternalMSEEDError:
                self._exceptions['InternalMSEEDError'] += 1
            except FDSNException:
                time.sleep(self.sleep_time_after_exception)
            except IncompleteRead:
                self._exceptions['IncompleteRead'] += 1
                
        return None
    
    
    def handle_multiple_locations(self, st, station_info):
        """ Automatic selection of location.
        
        Parameters
        ----------
        st : obspy.core.stream.Stream
        
        station_info : obspy.core.inventory.station.Station
        
        
        Returns
        -------
        st : obspy.core.stream.Stream
        """
        locations = set([tr.stats.location for tr in st])
        if len(locations) > 1:
            channel_type = st[0].stats.channel[0]
            for channel in station_info:
                if channel.location_code not in locations:
                    continue
                if channel.code[0] != channel_type:
                    continue
                if 'borehole' in str(channel.sensor.description).lower():
                    location = channel.location_code
                    break
            else:
                location = sorted(set([tr.stats.location for tr in st]))[0]    
            return st.select(location=location)
        
        return st


    def prepare_data(self, st):
        """ Demean, detrend, tapering, removing response and resampling
        
        Parameters
        ----------
        st : obspy.core.stream.Stream
        
        
        Returns
        -------
        obspy.core.stream.Stream if the processing is successful, else None
        """
        try:
            st.detrend('demean')
            st.detrend('linear')
        except:
            self._exceptions['detrend'] += 1
            return
        st.taper(type='hann', max_percentage=0.05)        
        inventory = None if self.attach_response else self.inventory
        try:
            st.remove_response(output=self.units, 
                               inventory=inventory,
                               pre_filt=self.prefilter)
        except ValueError:
            self._exceptions['Response error'] += 1
            return
            
        sampling_rates = set([tr.stats.sampling_rate for tr in st])
        if len(sampling_rates) > 1:
            self._exceptions['Multiple sampling rates'] += 1
            return
        if sampling_rates.pop() != self.sampling_rate:
            st = resample(st, self.sampling_rate)
        return st


    def adjust_channels(self, st):
        """
        If the stream contains the Z12 channels, these are rotated towards ZNE.
        
        Parameters
        ----------
        st : obspy.core.stream.Stream
        
        Returns
        -------
        obspy.core.stream.Stream if the rotation is successful, else None
        """
        channels = set([tr.stats.channel[-1] for tr in st])
        if 'Z' in channels and 'N' in channels and 'E' in channels:
            return st
        sorting_priorities = {'Z':0, '1':1, '2':2, '3':3}
        sorting_method = lambda i: sorting_priorities[i]
        try:
            channels = ''.join(sorted(channels, key=sorting_method))
        except KeyError:
            self._exceptions['rotation (unknown orientation)'] += 1
        try:
            st = rotate_stream(st, method='->ZNE', inventory=self.inventory, 
                               components=[channels])
        except Exception:
            self._exceptions['rotation'] += 1
            return
        channels = [tr.stats.channel[-1] for tr in st]
        assert 'Z' in channels and 'N' in channels and 'E' in channels
        return st


    def select_components(self, st, baz):
        """ Handles the absence of some components in the final stream
        
        Parameters
        ----------
        st : obspy.Stream
        
        baz : int, float (in degrees)
            Back-azimuth used for the rotation NE->RT
        
        Returns
        -------
        st : obspy.core.stream.Stream
            If the expected three-component stream lacks of one or both the
            horizontal components, the vertical component is returned. If the
            stream lacks the vertical component but has the two horizontal ones,
            it returns the horizontal components rotated towards the back
            azimuth. If only one horizontal component is available, it returns 
            None        
        """
        channels = [tr.stats.channel[-1] for tr in st]
        if 'Z' in channels:
            return st.select(channel='*Z')
        if 'E' in channels and 'N' in channels:
            assert len(st) == 2
            try:
                st = rotate_stream(st, method='NE->RT', back_azimuth=baz)
            except ValueError:
                self._exceptions['rotation'] += 1
                return None
            return st


    def preprocessing(self, st, station, baz):
        """ Preprocessing of the obspy stream
        
        The function calls sequentially the methods :meth:`handle_multiple_locations`,
        :meth:`prepare_data`, and :meth:`adjust_channels`.
        
        Parameters
        ----------
        st : obspy.core.stream.Stream
        
        station : obspy.core.inventory.station.Station
        
        baz : float
            Back azimuth (degrees) of the epicenter with respect to the receiver
        
        Returns
        -------
        obspy.core.stream.Stream if the preprocessing is successful, else None
        """
        if st is None or len(st)==0:# or len(st)<self.components:
            return
        st = self.handle_multiple_locations(st, station)
        st = self.prepare_data(st)
        if st is None or len(st)>self.components:# or len(st)!=self.components:
            return
        
        if len(st) < self.components:
            return self.select_components(st, baz)
        
        elif len(st) == self.components == 3:
            st = self.adjust_channels(st)
            if st is None:
                return
            try:
                return rotate_stream(st, method='NE->RT', back_azimuth=baz)
            except ValueError:
                self._exceptions['rotation'] += 1
                return
            
        return st
    
    
    def compile_header_and_save(self, st, savedir, stla, stlo, stel, evla, evlo, evdp, 
                                otime, mag, dist, az, baz):
        """ Compiles the header of the obspy stream (sac format) and writes to disk
        
        Parameters
        ----------
        stream : obspy.core.stream.Stream
        
        stla, stlo, stel : float
            Latitude, longitude, elevation of the seismic station
            
        evla, evlo, evdp : float
            Latitude, longitude, depth of the earthquake
            
        otime : obspy.core.utcdatetime.UTCDateTime
            Origin time of the earthquake
            
        mag, dist : float
            Event magnitude and distance (km) of the event from the receiver
            
        az, baz : float
            Azimuth and back azimuth of the epicenter with respect to the receiver
        """
        for tr in st:
            tr.stats.sac = AttribDict()
            tr.stats.sac.stla = stla
            tr.stats.sac.stlo = stlo
            tr.stats.sac.stel = stel
            tr.stats.sac.nzhour = tr.stats.starttime.hour
            tr.stats.sac.nzjday = tr.stats.starttime.julday
            tr.stats.sac.nzmin = tr.stats.starttime.minute
            tr.stats.sac.nzmsec = tr.stats.starttime.microsecond
            tr.stats.sac.nzsec = tr.stats.starttime.second
            tr.stats.sac.nzyear = tr.stats.starttime.year
            tr.stats.sac.e = tr.stats.endtime - tr.stats.starttime
            tr.stats.sac.evla = evla
            tr.stats.sac.evlo = evlo
            tr.stats.sac.evdp = evdp
            tr.stats.sac.mag = mag
            tr.stats.sac.o = otime.timestamp - tr.stats.starttime.timestamp
            tr.stats.sac.dist = dist
            tr.stats.sac.az = az
            tr.stats.sac.baz = baz
            outfile = os.path.join(savedir, str(otime.timestamp), '%s.sac'%tr.id)
            tr.write(outfile, format='sac')
        
    
    def plot_stations(self, ax=None, show=True, oceans_color='water', 
                      lands_color='land', edgecolor='k', projection='Mercator',
                      resolution='110m', color_by_network=True, legend_dict={}, 
                      **kwargs):
        """ Creates a maps of seismic receivers available for download
        
        Parameters
        ----------
        stations : dict
            Dictionary object containing stations information. This should 
            structured so that each key corresponds to a station code 
            ($network_code.$station_code) and each value is a tuple containing 
            latitude and longitude of the station. 
            
            For example::
                
                { net1.sta1 : (lat1, lon1),
                  net1.sta2 : (lat2, lon2),
                  net2.sta3 : (lat3, lon3)
                  }
        
        ax : cartopy.mpl.geoaxes.GeoAxesSubplot
            If not None, the receivers are plotted on the `GeoAxesSubplot` instance. 
            Otherwise, a new figure and `GeoAxesSubplot` instance is created
            
        show : bool
            If True, the plot is shown. Otherwise, a `GeoAxesSubplot` instance is
            returned. Default is `True`
            
        oceans_color, lands_color : str
            Color of oceans and lands. The arguments are ignored if ax is not
            `None`. Otherwise, they are passed to `cartopy.feature.NaturalEarthFeature` 
            (to the argument 'facecolor'). Defaults are 'water' and 'land'
            
        edgecolor : str
            Color of the boundaries between, e.g., lakes and land. The argument 
            is ignored if ax is not None. Otherwise, it is passed to 
            `cartopy.feature.NaturalEarthFeature` (to the argument 'edgecolor'). 
            Default is 'k' (black)
            
        projection : str
            Name of the geographic projection used to create the `GeoAxesSubplot`.
            (Visit the `cartopy website 
            <https://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html>`_ 
            for a list of valid projection names.) If ax is not None, `projection` 
            is ignored. Default is 'Mercator'
        
        resolution : {'10m', '50m', '110m'}
            Resolution of the Earth features displayed in the figure. Passed to
            `cartopy.feature.NaturalEarthFeature`. Valid arguments are '110m',
            '50m', '10m'. Default is '110m'
        
        color_by_network : bool
            If `True`, each seismic network will have a different color in the
            resulting map, and a legend will be displayed. Otherwise, all
            stations will have the same color. Default is `True`
        
        legend_dict : dict, optional
            Dictionary of keyword arguments passed to `matplotlib.pyplot.legend`
        
        **kwargs : dict, optional
            Additional keyword arguments passed to `matplotlib.pyplot.scatter` 
            
        Returns
        -------
        If `show` is True, None, else `ax`, i.e. the GeoAxesSubplot
        """   
        stations = {}
        for net, sta in self.inventory_iterator(self.inventory):
            station_code = '%s.%s'%(net.code, sta.code)
            stla, stlo, stel = self.station_coordinates(sta)
            stations[station_code] = (stla, stlo)
        
        return plot_stations(stations=stations,
                             ax=ax, 
                             show=show, 
                             oceans_color=oceans_color, 
                             lands_color=lands_color, 
                             edgecolor=edgecolor, 
                             projection=projection,
                             resolution=resolution,
                             color_by_network=color_by_network, 
                             legend_dict=legend_dict,
                             **kwargs) 
            
    
    
    def start(self):
        """ Starts the downloads.
        
        The catalog of seismic events is iterated over and, for each event, all
        the waveforms from the receivers associated with `stations_config`
        will be downloaded, if available. The waveforms are detrended, demeaned, 
        removed of the instrument response, and resampled. Finally, they are
        saved in savedir/data/event, where 'event' is the origin time (in the
        obspy timestamp format) of the considered earthquake.
        
        The algorithm keeps track of the progress made. This allows one to stop
        the downloads and start from where it was left off at any time.
        """
        
        def get_done_stations(savedir, event):
            done_file = os.path.join(savedir, event, 'DONE.txt')
            if not os.path.exists(done_file):
                return set()
            return set([i.strip() for i in open(done_file)])
        
        def get_starttime(savedir, inv):
            remove_file(os.path.join(savedir, '.DS_STORE')) # mac tmp file
            events = sorted(os.listdir(savedir), key=lambda i: float(i))
            if not events:
                return None
            station_codes = set([
                    '%s.%s'%(net.code, sta.code) \
                    for net, sta in EQDownloader.inventory_iterator(inv)
            ])
            for event in events:
                done_stations = get_done_stations(savedir, event)
                if station_codes.difference(done_stations):
                    return UTC(float(event)) - 1
            return UTC(float(event)) + 1
        
        def update_done_file(file, station_code):
            with open(file, 'a') as f:
                f.write('%s\n'%station_code)
                
                
        savedir = os.path.join(self.savedir, 'data')
        os.makedirs(savedir, exist_ok=True)
        t1 = get_starttime(savedir, self.inventory)
        
        if t1 is None:
            t1 = self.starttime
        endtime = self.endtime
        while endtime > t1:
            t2 = t1+ONE_YEAR if t1+ONE_YEAR<=endtime else endtime
            
            catalog = self.fetch_catalog(t1, t2, **self.events_config)
            for event in catalog:
                (otime, mag), (evla, evlo, evdp) = self.get_event_info(event)
                outdir = os.path.join(savedir, str(otime.timestamp))
                os.makedirs(outdir, exist_ok=True)
                event.write(os.path.join(outdir, '%s.xml'%str(otime.timestamp)),
                            format='quakeml')
                done_file = os.path.join(savedir, str(otime.timestamp), 'DONE.txt')

                done_stations = get_done_stations(savedir, str(otime.timestamp))
                if self.verbose:
                    print('EVENT:', otime.isoformat().split('T')[0])
                    print('- Mag: %s\n- Lat: %.2f\n- Lon: %.2f'%(mag, evla, evlo))
                for net, sta in self.inventory_iterator(self.inventory):
                    station_code = '%s.%s'%(net.code, sta.code)
                    if station_code in done_stations:
                        continue
                    if not self.station_was_active(sta, otime):
                        update_done_file(done_file, station_code)
                        continue
                    stla, stlo, stel = self.station_coordinates(sta)
                    dist, az, baz = gps2dist_azimuth(evla, evlo, stla, stlo)
                    dist /= 1000
                    if dist>self.distmax or dist<self.distmin:
                        update_done_file(done_file, station_code)
                        continue
                    channels = self.active_channels(sta)
                    st = self.collect_waveforms(network=net.code, 
                                                station=sta.code,
                                                channels=channels,
                                                starttime=otime + dist/self.vmax,
                                                endtime=otime + dist/self.vmin)
                    st = self.preprocessing(st, sta, baz)
                    if st is None:
                        update_done_file(done_file, station_code)
                        continue
                    self.compile_header_and_save(st=st, 
                                                 savedir=savedir, 
                                                 stla=stla, 
                                                 stlo=stlo, 
                                                 stel=stel, 
                                                 evla=evla, 
                                                 evlo=evlo, 
                                                 evdp=evdp, 
                                                 otime=otime, 
                                                 mag=mag, 
                                                 dist=dist, 
                                                 az=az, 
                                                 baz=baz)
                    update_done_file(done_file, station_code)
                    self._downloaded += self.components
                    if self.verbose:
                        print(station_code)
                
                t1 = t2
                self._events_done += 1
                if self.verbose:
                    print(self)            
                
            
        
        


















